@@ -241,7 +241,7 @@ async def test_stat_when_no_mlst(pair_factory):
 @pytest.mark.asyncio
 async def test_stat_mlst(pair_factory):
     async with pair_factory() as pair:
-        now = dt.datetime.utcnow()
+        now = dt.datetime.now(tz=dt.timezone.utc).replace(tzinfo=None)
         await pair.make_server_files("foo")
         info = await pair.client.stat("foo")
         assert info["type"] == "file"
