@@ -6,14 +6,18 @@ if sys.version_info < (3, 11):
     pytest.skip(reason="required python 3.11+", allow_module_level=True)
 
 
-async def _auth_response(connection, rest):
-    connection.response("234", ":P")
-    return True
+def simple_response(code, message):
+    async def command(connection, rest):
+        connection.response(code, message)
+        return True
+
+    return command
 
 
-async def _ok_response(connection, rest):
-    connection.response("200", ":P")
-    return True
+AUTH_RESPONSE = simple_response("234", ":P")
+PBSZ_RESPONSE = simple_response("200", ":P")
+PROT_RESPONSE = simple_response("200", ":P")
+OK_RESPONSE = simple_response("200", ":P")
 
 
 @pytest.mark.asyncio
@@ -22,7 +26,9 @@ async def test_upgrade_to_tls(mocker, pair_factory):
     create_default_context = mocker.patch("aioftp.client.ssl.create_default_context", return_value=ssl_context)
 
     async with pair_factory(logged=False, do_quit=False) as pair:
-        pair.server.commands_mapping["auth"] = _auth_response
+        pair.server.commands_mapping["auth"] = AUTH_RESPONSE
+        pair.server.commands_mapping["pbsz"] = PBSZ_RESPONSE
+        pair.server.commands_mapping["prot"] = PROT_RESPONSE
 
         start_tls = mocker.patch.object(pair.client.stream, "start_tls")
         command_spy = mocker.spy(pair.client, "command")
@@ -31,7 +37,14 @@ async def test_upgrade_to_tls(mocker, pair_factory):
 
     create_default_context.assert_called_once_with()
     start_tls.assert_called_once_with(sslcontext=ssl_context, server_hostname=pair.client.server_host)
-    command_spy.assert_called_once_with("AUTH TLS", "234")
+    assert command_spy.call_count == 3
+    command_spy.assert_has_calls(
+        [
+            mocker.call("AUTH TLS", "234"),
+            mocker.call("PBSZ 0", "200"),
+            mocker.call("PROT P", "200"),
+        ],
+    )
 
 
 @pytest.mark.asyncio
@@ -40,7 +53,9 @@ async def test_upgrade_to_tls_custom_ssl_context(mocker, pair_factory):
     create_default_context = mocker.patch("aioftp.client.ssl.create_default_context")
 
     async with pair_factory(logged=False, do_quit=False) as pair:
-        pair.server.commands_mapping["auth"] = _auth_response
+        pair.server.commands_mapping["auth"] = AUTH_RESPONSE
+        pair.server.commands_mapping["pbsz"] = PBSZ_RESPONSE
+        pair.server.commands_mapping["prot"] = PROT_RESPONSE
 
         start_tls = mocker.patch.object(pair.client.stream, "start_tls")
         command_spy = mocker.spy(pair.client, "command")
@@ -49,23 +64,14 @@ async def test_upgrade_to_tls_custom_ssl_context(mocker, pair_factory):
 
     create_default_context.assert_not_called()
     start_tls.assert_called_once_with(sslcontext=ssl_context, server_hostname=pair.client.server_host)
-    command_spy.assert_called_once_with("AUTH TLS", "234")
-
-
-@pytest.mark.asyncio
-async def test_upgrade_to_tls_does_nothing_when_already_updated(mocker, pair_factory):
-    mocker.patch("aioftp.client.ssl.create_default_context")
-
-    async with pair_factory(logged=False, do_quit=False) as pair:
-        pair.client._upgraded_to_tls = True
-
-        start_tls = mocker.patch.object(pair.client.stream, "start_tls")
-        command_spy = mocker.spy(pair.client, "command")
-
-        await pair.client.upgrade_to_tls()
-
-    start_tls.assert_not_called()
-    command_spy.assert_not_called()
+    assert command_spy.call_count == 3
+    command_spy.assert_has_calls(
+        [
+            mocker.call("AUTH TLS", "234"),
+            mocker.call("PBSZ 0", "200"),
+            mocker.call("PROT P", "200"),
+        ],
+    )
 
 
 @pytest.mark.asyncio
@@ -73,10 +79,9 @@ async def test_upgrade_to_tls_when_logged_in(mocker, pair_factory):
     mocker.patch("aioftp.client.ssl.create_default_context")
 
     async with pair_factory(logged=False, do_quit=False) as pair:
-        pair.server.commands_mapping["auth"] = _auth_response
-        pair.server.commands_mapping["pbsz"] = _ok_response
-        pair.server.commands_mapping["prot"] = _ok_response
-        pair.client._logged_in = True
+        pair.server.commands_mapping["auth"] = AUTH_RESPONSE
+        pair.server.commands_mapping["pbsz"] = OK_RESPONSE
+        pair.server.commands_mapping["prot"] = OK_RESPONSE
 
         mocker.patch.object(pair.client.stream, "start_tls")
         command_spy = mocker.spy(pair.client, "command")
@@ -90,24 +95,3 @@ async def test_upgrade_to_tls_when_logged_in(mocker, pair_factory):
             mocker.call("PROT P", "200"),
         ],
     )
-
-
-@pytest.mark.asyncio
-async def test_login_should_send_tls_protection_when_upgraded(mocker, pair_factory):
-    mocker.patch("aioftp.client.ssl.create_default_context")
-
-    async with pair_factory(logged=False, do_quit=False) as pair:
-        pair.server.commands_mapping["pbsz"] = _ok_response
-        pair.server.commands_mapping["prot"] = _ok_response
-        pair.client._upgraded_to_tls = True
-
-        command_spy = mocker.spy(pair.client, "command")
-
-        await pair.client.login("foo", "bar")
-
-    command_spy.assert_has_calls(
-        [
-            mocker.call("PBSZ 0", "200"),
-            mocker.call("PROT P", "200"),
-        ],
-    )
